'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import random
import os

# Selenium construct. Needs to create wrapper exception and use in dispatcher.
from selenium.common.exceptions import StaleElementReferenceException


from arjuna.interact.gui.auto.base.locatable import Locatable
from arjuna.interact.gui.auto.base.dispatchable import Dispatchable
from arjuna.interact.gui.auto.base.configurable import Configurable
from arjuna.interact.gui.auto.element.guielement import GuiElement
from arjuna.interact.gui.auto.source.parser import *

class _GuiPartialElement(GuiElement):

    def __init__(self, gui, multi_element, index: int, dispatcher_element, iconfig=None):
        super().__init__(gui, multi_element.lmd, iconfig=iconfig)
        self.__multi_element = multi_element
        self.__index = index
        self.dispatcher = dispatcher_element
        self.load_source_parser()

    def create_dispatcher(self):
        pass

    def _is_partial_element(self):
        return True

    @property
    def index(self):
        return self.__index

class GuiMultiElement(Locatable,Dispatchable,Configurable):
    
    def __init__(self, gui, lmd, iconfig=None, elements=None): #, parent=None):
        Locatable.__init__(self, gui, lmd) #, parent)
        Dispatchable.__init__(self)
        Configurable.__init__(self, gui, iconfig)
        if elements:
            self.__elements = elements
        else:
            self.__elements = list()   
        self.__source_parser = None

    # def configure_partial_elements(self, elem_config):
    #     '''
    #         This method is supposed to be called when multielement identification is completed.
    #         This is not used for usual multi-element.
    #         It is used by RadioGroup or DropDown etc which are higher level abstractions that use ME.
    #     '''
    #     for instance in self.__elements:
    #         instance.configure(elem_config)

    # def find(self):
    #     self.parent_container.find_multielement(self)
    #     self.__load_source_parser()

    # #Override
    # def find_if_not_found(self):
    #     if not self.is_found():
    #         self.find()

    def load_source_parser(self):
        self.__source_parser = MultiElementSource()
        self.__source_parser.load(self.__elements)

    def __getitem__(self, index):
        # self.find_if_not_found()
        return self.__elements[index]

    @property
    def size(self):
        # self.find_if_not_found()
        return len(self.__elements)

    @size.setter
    def size(self, count):
        # The logic ignores stale elements 
        for i in range(count):
            try:
                e = _GuiPartialElement(self.gui, self, i, self.dispatcher.get_element_at_index(i), iconfig=self.settings)
                self.__elements.append(e)
            except StaleElementReferenceException as e:
                pass

    length = size

    @property
    def random_element(self):
        # self.find_if_not_found()
        return self[random.randint(0, self.size-1)]

    @property
    def first_element(self):
        return self[0]

    @property
    def last_element(self):
        return self[self.length-1]

    def get_element_at_ordinal(self, ordinal):
        # self.find_if_not_found()
        return self.__elements[ordinal-1]

    def get_instance_by_visible_text(self, text):
        texts = self.__get_all_texts()
        first_index = self.__find_first_text_index(texts, text)
        return self[first_index]

    def get_instance_by_value(self, value):
        values = self.__get_all_values()
        first_index = self.__find_first_value_index(values, value)
        return self[first_index]

    def wait_until_visible(self):
        self.find_if_not_found()
        # Should the logic be keptin Setu itself for this????

    def __return_attr_values(self, response):
        if "data" in response and "attrValues" in response["data"]:
            return response["data"]["attrValues"]
        else:
            return None

    def are_selected(self):
        return [instance.is_selected() for instance in self.__elements]

    # getting index attribute when it does not exist retursn value attribute.
    # So, not going the Selenium way. Setu would treat index as computer counting.
    def has_index_selected(self, index):
        # self.find_if_not_found()
        return self[index].is_selected()

    # Ordinal is human counting
    def has_ordinal_selected(self, ordinal):
        return self.has_index_selected(ordinal-1)

    def __find_first_match_index(self, in_sequence, to_match):
        try:
            return in_sequence.index(to_match)
        except:
            return -1

    def __get_all_texts(self):
        # self.find_if_not_found()
        self.load_source_parser()
        texts = self.__source_parser.get_text_contents()
        return texts

    def __find_first_text_index(self, texts, text):
        first_index = texts.index(text)
        if first_index == -1:
            raise Exception("No option with {} visible text present in drop down.".format(text))
        return first_index

    def __get_all_values(self):
        # self.find_if_not_found()
        self.load_source_parser()
        values = self.__source_parser.get_values()
        return values

    def __find_first_value_index(self, values, value):
        first_index = values.index(value)
        if first_index == -1:
            raise Exception("No option with {} value present in drop down.".format(value))
        return first_index

    def get_first_selected_instance(self):
        # self.find_if_not_found()
        self.load_source_parser()
        booleans = self.are_selected()
        first_index = None
        try:
            first_index = booleans.index(True)
        except:
            raise Exception("No option in drop down is currenlty selected.")
        else:
            return self[first_index]

    @property
    def source(self):
        return self.__source_parser

    @property
    def elements(self):
        return self.__elements

    @property
    def filter(self):
        return ElementFilter(self)

class ElementFilter:

    def __init__(self, gui_multi_element):
        self.__gui = gui_multi_element.gui
        self.__lmd = gui_multi_element.lmd
        self.__iconfig = gui_multi_element.settings
        self.__elements = gui_multi_element.elements
        self.__filtered_elements = self.__elements

    def build(self):
        me = GuiMultiElement(self.__gui, self.__lmd, iconfig=self.__iconfig, elements=self.__filtered_elements)
        self.__filtered_elements = self.__elements
        return me

    def active(self):
        out = list()
        for e in self.__filtered_elements:
            try:
                e.get_html()
                out.append(e)
            except StaleElementReferenceException as e:
                pass
        self.__filtered_elements = out
        return self

    def visible(self):
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.is_visible()
        ]
        return self

    def attr(self, name):
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.source.is_attr_present(name)
        ]
        return self        

    def attr_value(self, name, value):
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.source.get_attr_value(name) == value
        ]
        return self

    def value(self, value):
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.source.get_value("value") == str(value)
        ]
        return self

    def text_content(self, text):
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if text in e.text
        ]
        return self     

    


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import traceback
from arjuna.interact.gui.auto.element.guielement import GuiElement
from arjuna.core.enums import ArjunaOption
from arjuna.interact.gui.auto.finder.emd import SimpleGuiElementMetaData

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.core.exceptions import ChildWindowNotFoundError

class WindowConditions:

    def __init__(self, window):
        self.__window = window

    @property
    def window(self):
        return self.__window

    def ChildWindowIsPresent(self, lmd):
        caller = DynamicCaller(self.window._find_child_window, lmd)
        return CommandCondition(caller)

class BasicWindow:

    def __init__(self, app, automator):
        super().__init__()
        self.__app = app
        self.__automator = automator
        self.__window_handle = None
        self.__config = automator.config

    @property
    def max_wait_time(self):
        return self.automator.config.guiauto_max_wait_time

    @property
    def app(self):
        return self.__app

    @property
    def config(self):
        return self.__config    

    @property
    def automator(self):
        return self.__automator

    @property
    def handle(self):
        return self.__window_handle

    @property
    def title(self):
        return self.automator.dispatcher.get_current_window_title()

    @property
    def size(self):
        return self.automator.dispatcher.get_current_window_size()

    def _set_handle(self, handle):
        self.__window_handle = handle

    def focus(self):
        self.automator.dispatcher.focus_on_window(self.handle)

    def is_main_window(self):
        return False

    def set_window_size(self, width, height):
        self.automator.dispatcher.set_current_window_size(width, height)

    def maximize(self):
        self.automator.dispatcher.maximize_current_window()

class MainWindow(BasicWindow):

    def __init__(self, app, automator):
        super().__init__(app, automator)
        self._set_handle(self.get_current_window_handle()) 
        self.__all_child_windows = {}
        self.__setu_id_map = {}
        self.__resize_window_as_per_config()
        self.__conditions = WindowConditions(self)

    @property
    def conditions(self):
        return self.__conditions

    def get_all_child_window_handles(self):
        handles = self.automator.dispatcher.get_all_window_handles()
        handles = [handle for handle in handles if handle != self.handle]
        new_handles = []
        for handle in handles:
            if handle != self.handle:
                if handle not in self.__all_child_windows:
                    cwin = ChildWindow(self.app, self.automator, self, handle)
                    self.__all_child_windows[handle] = cwin
                    new_handles.append(handle)
        return handles, new_handles

    @property
    def latest_child_window(self):
        _, new_handles = self.get_all_child_window_handles()
        if not new_handles:
            raise Exception("No new window was launched.")
        elif len(new_handles) > 1:
            raise Exception("Multiple new windows were launched, so can not deterministically jump to a new window.")
        return self.__all_child_windows[new_handles[0]]

    def __get_child_window(self, handle):
        return self.__all_child_windows[handle]

    def __focus_on_window(self, handle):
        self.__get_child_window(handle).jump()

    def delete_window(self, handle):
        del self.__all_child_windows[handle]

    def close_all_child_windows(self):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            cwin.close()
        self.focus()

    def get_current_window_handle(self):
        return self.automator.dispatcher.get_current_window_handle()

    def child_window(self, locator_meta_data):
        return self.conditions.ChildWindowIsPresent(locator_meta_data).wait(max_wait_time=self.max_wait_time)

    def _find_child_window(self, locator_meta_data):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            for locator in locator_meta_data.locators:
                if locator.ltype.name == "WINDOW_TITLE":
                    if cwin.title.lower() == locator.lvalue.lower():
                        return cwin
                elif locator.ltype.name == "WINDOW_PTITLE":
                    if locator.lvalue.lower() in cwin.title.lower():
                        return cwin
                elif locator.ltype.name == "CONTENT_LOCATOR":
                    for child_locator in locator.lvalue.locators:
                        try:
                            emd = SimpleGuiElementMetaData(child_locator.ltype.name, child_locator.lvalue)
                            # The element for window is created in the context of an app.
                            # Need to verify this logic.
                            # and its impact on POM.
                            contained_element = self.automator.element(self.app,emd, max_wait_time=0.5)
                            # contained_element.find()
                            return cwin
                        except WaitableError as f:
                            continue  
                        except Exception as e:
                            raise Exception(str(e) + traceback.format_exc())                          
        
        raise ChildWindowNotFoundError(*locator_meta_data.locators)

    def __resize_window_as_per_config(self):
        # Resize window
        config = self.config
        browser_width = config.get_arjuna_option_value(ArjunaOption.BROWSER_DIM_WIDTH)
        browser_height = config.get_arjuna_option_value(ArjunaOption.BROWSER_DIM_HEIGHT)
        should_maximize = config.get_arjuna_option_value(ArjunaOption.BROWSER_MAXIMIZE)

        if config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_HEIGHT):
            if should_maximize:
                self.maximize()
        else:
            width, height = None, None
            current_width, current_height = self.get_size()
            width = config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and browser_width or current_width
            height = config.is_arjuna_option_not_set(ArjunaOption.BROWSER_DIM_HEIGHT) and browser_height or current_height
            self.set_window_size(width, height)

    def is_main_window(self):
        return True

    def close(self):
        raise Exception("You can not close main window. Use automator.quit() to quit application.")

class ChildWindow(BasicWindow):

    def __init__(self, app, automator, main_window, handle):
        super().__init__(app, automator)
        self.__main_window = main_window
        self._set_handle(handle) 

    def close(self):
        self.focus()
        self.automator.dispatcher.close_current_window()
        self.__main_window.delete_window(self.handle)
        self.__main_window.focus()


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .gui import *

class Page(AppContent):

    def __init__(self, *args, source_gui, label=None, **kwargs):
        # app = isinstance(source_gui, App) and source_gui or source_gui.app
        super().__init__(automator=source_gui.automator, label=label)
        self.app.ui = self
        self._load(*args, **kwargs)

class Widget(AppContent):

    def __init__(self, page, *args, label=None, **kwargs):
        super().__init__(automator=page.automator, label=label, page=page)   
        self.__page = page
        self._load(*args, **kwargs)

    @property
    def page(self):
        return self.__page

class App(Gui, metaclass=abc.ABCMeta):

    def __init__(self, *, config=None, ext_config=None, label=None, ns_dir=None):
        super().__init__(config=config, ext_config=ext_config, label=label)
        self.__ui = None
        self.__automator = None
        self.__ns_dir = ns_dir

    @property
    def automator(self):
        return self.__automator

    @property
    def ns_dir(self):
        return self.__ns_dir

    def _launchautomator(self):
        # Default Gui automation engine is Selenium
        from arjuna.interact.gui.auto.automator import GuiAutomator
        self.__automator = GuiAutomator(self, self.config, self.ext_config)    

    @property
    def ui(self):
        return self.__ui

    @ui.setter
    def ui(self, page):
        self.__ui = page

    def _create_default_ui(self):
        self.__ui = Page(source_gui=self, label="{}-Def-UI".format(self.label))

    @abc.abstractmethod
    def launch(self):
        pass

    def prepare_widget(self, widget_object):
        return widget_object


class WebApp(App):

    def __init__(self, *, base_url=None, blank_slate=False, config=None, ext_config=None, label=None, ns_dir=None):
        '''
            Creates and returns GuiAutomator object for provided config.
            If no configuration is provided reference configuration is used.
            You can also provide GuiDriverExtendedConfig for extended configuration for WebDriver family of libs. 
        '''
        super().__init__(config=config, ext_config=ext_config, label=label, ns_dir=ns_dir)
        from arjuna.core.enums import ArjunaOption
        self.__base_url = base_url is not None and base_url or self.config.get_arjuna_option_value(ArjunaOption.AUT_BASE_URL).as_str()
        # self._load(*args, **kwargs)

    @property
    def base_url(self):
        return self.__base_url

    def launch(self, blank_slate=False):
        self._launchautomator()
        if not blank_slate:
            self.automator.browser.go_to_url(self.base_url)
        self._create_default_ui()

    def quit(self):
        self.automator.quit()


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import sys
import os
import argparse
from functools import partial

from .validation import *
from arjuna.core.enums import ReportFormat


class Parser:

    def __init__(self):
        self.parser = None

    def get_parser(self):
        return self.parser

class ProjectParser(Parser):

    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        #self.parser.add_argument('-w', '--workspace-dir', dest="workspace_dir", type=str, help='Workspace Directory')
        self.parser.add_argument('-p', '--project-root-dir', dest="project.root.dir", type=project_dir, help = 'Valid absolute Project root directory. Name of project (Alnum 3-30 length. Only lower case letters.).')

    def process(self, arg_dict):
        # arg_dict['workspace.dir'] = arg_dict['workspace_dir']
        # del arg_dict['workspace_dir']
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class NewProjectParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        #self.parser.add_argument('-w', '--workspace-dir', dest="workspace_dir", type=str, help='Workspace Directory')
        self.parser.add_argument('-p', '--project-root-dir', dest="project.root.dir", type=new_project_dir, help = 'Absolute non-existing Project root directory. Name of project (Alnum 3-30 length. Only lower case letters.).')

    def process(self, arg_dict):
        # arg_dict['workspace.dir'] = arg_dict['workspace_dir']
        # del arg_dict['workspace_dir']
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class RunParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-rid", "--runid", dest="run.id", type=partial(lname_check, "Run ID"), help = 'Alnum 3-30 length. Only lower case letters.', default="mrun")
        self.parser.add_argument('-rf', '--report-formats', dest="project.report.formats", type=report_format, metavar=('F1','F2'), default=['XML', 'HTML'], nargs='+', help='One or more report format names.') # choices=['XML', 'HTML'], 
        self.parser.add_argument('-e', '--enumerate-only', dest="enumerate", action='store_true', help = 'Show picked tests without running them.')

        # self.parser.add_argument('-aco', '--arjuna-ref-option', dest="aro",
        #                          nargs=2,
        #                          action='append',
        #                          metavar=('option', 'value'),
        #                          help='Arjuna Reference Option. Can pass any number of these switches as -aco x y -aco z t')
        #
        # self.parser.add_argument('-ato', '--arjuna-ext-option', dest="aeo",
        #                          nargs=2,
        #                          action='append',
        #                          metavar=('option', 'value'),
        #                          help='Arjuna Extended Option. Can pass any number of these switches as -aco x y -ato z t')
        #
        # self.parser.add_argument('-uco', '--user-ref-option', dest="uro",
        #                          nargs=2,
        #                          action='append',
        #                          metavar=('option', 'value'),
        #                          help='User Reference Option. Can pass any number of these switches as -aco x y -uco z t')
        #
        # self.parser.add_argument('-uto', '--user-ext-option', dest="ueo",
        #                          nargs=2,
        #                          action='append',
        #                          metavar=('option', 'value'),
        #                          help='User Extended Option. Can pass any number of these switches as -aco x y -uto z t')

    def process(self, arg_dict):
        pass

class PickersParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-cm', '--cmodules', dest="cmodules", metavar=('M1','M2'), default=None, nargs='+', help='One or more names/patterns for considering test modules.')
        self.parser.add_argument('-im', '--imodules', dest="imodules", metavar=('M1','M2'), default=None, nargs='+', help='One or more names/patterns for ignoring test modules.')
        self.parser.add_argument('-cc', '--cclasses', dest="cclasses", metavar=('C1','C2'), default=None, nargs='+', help='One or more names/patterns for considering test classes.')
        self.parser.add_argument('-ic', '--iclasses', dest="iclasses", metavar=('C1','C2'), default=None, nargs='+', help='One or more names/patterns for ignoring test classes.')
        self.parser.add_argument('-cf', '--cfunctions', dest="cfunctions", metavar=('F1','F2'), default=None, nargs='+', help='One or more names/patterns for considering test functions/methods.')
        self.parser.add_argument('-if', '--ifunctions', dest="ifunctions", metavar=('F1','F2'), default=None, nargs='+', help='One or more names/patterns for ignoring test function/methods.')

    def process(self, arg_dict):
        pass

# class NamesParser(Parser):
#     def __init__(self):
#         super().__init__()
#         self.parser = argparse.ArgumentParser(add_help=False)
#         self.parser.add_argument('-cf', '--cfunctions', dest="cfunctions", metavar=('F1','F2'), default=None, nargs='+', help='One or more names/patterns for considering test functions.')
#         # self.parser.add_argument('-cm', '--cmodules', dest="cmodules", metavar=('M1','M2'), default=None, nargs='+', help='One or more names/patterns for considering test modules.')
#         # self.parser.add_argument('-im', '--imodules', dest="imodules", metavar=('M1','M2'), default=None, nargs='+',
#         #                  help='One or more names/patterns for ignoring test modules.')
#         # self.parser.add_argument('-cf', '--cfunctions', dest="cfunctions", metavar=('F1','F2'), default=None, nargs='+', help='One or more names/patterns for considering test functions.')
#         # self.parser.add_argument('-if', '--ifunctions', dest="ifunctions", metavar=('F1','F2'), default=None, nargs='+',
#         #                  help='One or more names/patterns for ignoring test functions.')

#     def process(self, arg_dict):
#         pass


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.interact.gui.auto.base.configurable import Configurable
from arjuna.interact.gui.auto.base.dispatchable import Dispatchable
from arjuna.interact.gui.auto.source.parser import ElementXMLSourceParser
from arjuna.core.exceptions import GuiElementTextNotSetError

class Interactable(Configurable, Dispatchable):

    def __init__(self, gui, iconfig): #, parent=None, find=True):
        self.__config = gui.automator.config
        Configurable.__init__(self, gui, iconfig)
        Dispatchable.__init__(self)
        self.__source = None

        #self.__automator = automator
        from arjuna.interact.gui.auto.condition.element_conditions import GuiElementConditions, GuiElementLenientInteraction
        self.__conditions_handler = GuiElementConditions(self)
        self.__interaction_handler = GuiElementLenientInteraction(self)
        self.__source_parser = None

        # # For partial element find is False
        # if find:
        #     self.find()

    @property
    def config(self):
        return self.__config

    def __get_attr_value_from_remote(self, attr, optional=False):
        return self.__return_attr_value(self.dispatcher.get_attr_value(attr, optional))

    def get_html(self):
        return self.__get_attr_value_from_remote("outerHTML")

    def load_source_parser(self):
        raw_source = self.get_html()
        self.__source_parser = ElementXMLSourceParser(raw_source)
        self.__source_parser.load()

    # def find(self):
    #     self.parent_container.find_element(self)
    #     self.load_source_parser()

    # def find_without_wait(self):
    #     self.parent_container.find_element_without_wait(self)

    # identify = find
    # reset = find

    def _is_partial_element(self):
        return False

    def _get_instance_number(self):
        raise Exception("Instance number is applicable only for partial gui elements.")

    # def _kwargs(self, **kwargs):
    #     return self.__append_instance_number(kwargs)

    # def _noargs(self):
    #     return self.__append_instance_number({})

    def _only_send_text(self, text):
        self.dispatcher.send_text(text)

    def _only_set_text(self, text):
        self._only_clear_text()
        self._only_send_text(text)
        entered_text = self.dispatcher.get_attr_value("value")
        if entered_text != text:
            raise GuiElementTextNotSetError("Expected text: {}. Actual text is {}".format(text, entered_text))

    def _only_click(self):
        # if self._should_scroll_to_view():
        #     self.dispatcher.scroll_to_view()
        self.dispatcher.click()

    def __return_attr_value(self, result):
        return result and result or None

    def scroll_full_height(self):
        self.dispatcher.scroll_full_height()

    def click(self):
        # self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self.interactions.Click().wait()
        #self._only_click()

    def __only_hover(self):
        self.dispatcher.hover()

    def hover(self):
        self.__wait_until_clickable_if_configured()
        self.__only_hover()

    def hover_and_click(self):
        self.__only_hover()
        self.__wait_until_clickable_if_configured()
        self.dispatcher.mouse_click()

    def __conditional_selected_state_click(self, condition_state):
        # self.find_if_not_found()
        selected = self.is_selected()
        if selected == condition_state:
            self.__wait_until_clickable_if_configured()
            self._only_click()

    def select(self):
        self.__conditional_selected_state_click(False)

    def deselect(self):
        self.__conditional_selected_state_click(True)

    # def reload(self):
    #     pass

    # wait_until_present = reload

    def wait_until_visible(self):
        # self.find_if_not_found()
        self.conditions.IsVisible().wait()

    def wait_until_clickable(self):
        # self.find_if_not_found()
        self.conditions.IsClickable().wait()

    def wait_until_selected(self):
        # self.find_if_not_found()
        self.conditions.IsVisible().wait()

    #################################
    ### State Checking
    #################################
    def is_selected(self):
        # self.find_if_not_found()
        return self.dispatcher.is_selected()

    def is_visible(self):
        # self.find_if_not_found()
        return self.dispatcher.is_visible()

    def is_clickable(self):
        # self.find_if_not_found()
        return self.dispatcher.is_clickable()

    @property
    def conditions(self):
        return self.__conditions_handler

    @property
    def interactions(self):
        return self.__interaction_handler

    def __wait_until_clickable_if_configured(self):
        if self._should_check_pre_state(): self.wait_until_clickable()

    # Properties
    @property
    def source(self): # OLD: refind=True, reload=True
        return self.__source_parser

    @property
    def text(self):
        return self.source.content.text

    # Textbox abstraction
    @text.setter
    def text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()
        self.interactions.SetText(text).wait()

    def clear_text(self):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()

    def send_text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_enter_text(text)

    def enter_text(self, text):
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_enter_text(text)

    def _only_clear_text(self):
        self.dispatcher.clear_text()

    def _only_enter_text(self, text):
        self._only_send_text(text)

    def has_entered_text(self, text):
        pass

    def has_entered_partial_text(self, text):
        pass

    # Checkbox abstraction
    def check(self):
        self.select()

    def uncheck(self):
        self.deselect()

    def toggle_checkbox(self):
        self.click()

    def is_checked(self):
        return self.is_selected()

    def get_property_value(self, attr):
        # self.find_if_not_found()
        return self.__get_attr_value_from_remote(attr)

    def perform_action_chain(self, single_action_chain):
        from arjuna.interact.gui.auto.automator.actions import SingleActionChain
        action_chain = SingleActionChain(self.__automator, element=self)
        action_chain.perform(single_action_chain)



'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
from arjuna.core.exceptions import WaitableError, GuiElementNotFoundError, GuiElementPresentError

class ElementFinder:
    def __init__(self, container): #, obj_name=""):
        self.__container = container
        self.__config = container.config
        # self.__obj_name = obj_name

    @property
    def config(self):
        return self.__config

    @property
    def container(self):
        return self.__container

    @abc.abstractmethod
    def _create_element_flat_or_nested(self, lmd):
        pass

    @abc.abstractmethod
    def _create_multielement_flat_or_nested(self, lmd):
        pass

    def check_for_absence(self, dispatcher_call, lmd, context="ELEMENT"):
        try:
            self.find(dispatcher_call, lmd, context)
        except GuiElementNotFoundError as e:
            # This is expected
            pass
        else:
            raise GuiElementPresentError(*lmd.locators)

    def find(self, dispatcher_call, lmd, context="ELEMENT"):
        found = False
        js_call_name = context == "ELEMENT" and "find_element_with_js" or "find_multielement_with_js"
        js_call = getattr(self.container, js_call_name)
        locators = lmd.locators
        if context != "ELEMENT":
            if "POINT" in {l.ltype.name for l in locators}:
                raise ConditionException("With.POINT can be used only with GuiElement.")

        we = None
        for locator in locators: 
            try:
                if locator.ltype.name == "POINT":
                    # Assumption here is that this container is automator.
                    size, dispatcher = js_call("return document.elementFromPoint({}, {})".format(*locator.lvalue))
                elif locator.ltype.name == "JAVASCRIPT":
                    size, dispatcher = js_call(locator.lvalue)
                else:
                    size, dispatcher = dispatcher_call(locator.ltype.name, locator.lvalue)
                return locator.ltype.name, locator.lvalue, size, dispatcher
            except WaitableError as e:
                we = e
            except Exception as f:
                raise f
            else:
                we = None
        if not found:
            raise GuiElementNotFoundError(*lmd.locators)


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.interact.gui.auto.base.configurable import Configurable

class GuiWebRadioGroup(Configurable):

    def __init__(self, gui, lmd, parent=None):
        super().__init__(gui)
        self.__gui = gui
        self.__automator = gui.automator
        self.__finder = parent and parent or self.automator
        self.__radios = None
        self.__find(self.__finder, lmd)

    @property
    def gui(self):
        return self.__gui

    @property
    def automator(self):
        return self.__automator

    def __validate_radio_buttons(self, source):
        if [t for t in source.get_tag_names() if t.strip().lower() != 'input']:
            raise Exception("Not a valid radio group. Contains non-input elements.")
        if [t for t in source.get_attr_values("type") if t.strip().lower() != 'radio']:
            raise Exception("Not a valid radio group. Contains non-radio elements.")
        names = source.get_attr_values("name")
        if len(set(names)) != 1:
            raise Exception("Not a valid radio group. Contains radio elements belonging to different radio groups.")

    def __check_type_if_configured(self, tags):
        if self._should_check_type(): self.__validate_radio_buttons(tags)

    def __find(self, finder, lmd):
        # This would force the identification of partial elements in the wrapped multi-element.
        self.__radios = finder.multi_element(self.gui, lmd)
        self.__check_type_if_configured(self.source)
        self.__radios.configure_partial_elements(self.settings)

    def has_index_selected(self, index):
        return self.__radios[index].is_selected()

    def has_value_selected(self, value):
        return self.__radios.get_instance_by_value(value).is_selected()

    @property
    def value(self):
        instance = self.__radios.get_first_selected_instance()
        return instance.source.get_attr_value("value")

    def __select_option(self, option):
        option.select()
        if self._should_check_post_state() and not option.is_selected():
            raise Exception("The attempt to select the radio button was not successful.")

    def select_index(self, index):
        option = self.__radios[index]
        self.__select_option(option)

    def select_ordinal(self, ordinal):
        return self.select_by_index(ordinal-1)

    def select_value(self, value):
        option = self.__radios.get_instance_by_value(value)
        self.__select_option(option)

    @property
    def source(self):
        return self.__radios.source


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
import pytest

from arjuna.core.enums import ReportFormat

class TestRunner:
    import sys
    import os
    import unittest
    
    def __init__(self):
        from arjuna import Arjuna
        from arjuna.core.enums import ArjunaOption
        self.__project_dir = Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_ROOT_DIR).as_str()
        self.__tests_dir = Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_TESTS_DIR).as_str()
        self.__xml_path = os.path.join(Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_RUN_REPORT_XML_DIR).as_str(), "report.xml")
        self.__html_path = os.path.join(Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_RUN_REPORT_HTML_DIR).as_str(), "report.html")
        self.__report_formats = Arjuna.get_ref_config().get_arjuna_option_value(ArjunaOption.PROJECT_REPORT_FORMATS).as_enum_list(ReportFormat)
        # -s is to print to console.
        self.__pytest_args = ["--rootdir", self.__project_dir, "--no-print-logs", "-s"]
        self.__test_args = []

    @property
    def tests_dir(self):
        return self.__tests_dir

    @property
    def test_suite(self):
        return self.__suite
    
    def load_all_tests(self):
        self.__pytest_args.insert(0, self.tests_dir)

    def load_tests_from_pickers(self, *, cm=None, im=None, cc=None, ic=None, cfn=None, ifn=None):  

        def process_modules(ms):
            ms = [m.replace(".py", "").replace("*","").replace("/", " and ").replace("\\", " and ") for m in ms]
            return ["and" in m and "({})".format(m) or m for m in ms]

        k_args = []

        k_flag = False

        if im:            
            im = process_modules(im)
            k_args.append(" and ".join(["not " + m for m in im]))
            k_flag = True

        if ic:
            prefix = k_flag and " and " or ""
            k_args.append(prefix + " and ".join(["not " + c for c in ic]))
            k_flag = True

        if ifn:
            prefix = k_flag and " and " or ""
            k_args.append(prefix + " and ".join(["not " + c for c in ifn]))
            k_flag = True

        if cm:
            prefix = k_flag and " and " or ""            
            cm = process_modules(cm)
            k_args.append(prefix + " or ".join(cm))
            k_flag = True

        if cc:
            prefix = k_flag and " and " or "" 
            k_args.append(prefix + " or ".join(cc))
            k_flag = True

        if cfn:
            prefix = k_flag and " and " or "" 
            k_args.append(prefix + " or ".join(cfn))
            k_flag = True

        if k_flag:
            self.__test_args.append("-k " + "".join(k_args))
        
  
    def run(self, *, only_enumerate):
        from arjuna import Arjuna
        from arjuna.core.enums import ArjunaOption

        pytest_report_args = []

        if ReportFormat.XML in self.__report_formats:
            pytest_report_args.extend(["--junit-xml", self.__xml_path])

        if ReportFormat.HTML in self.__report_formats:
            pytest_report_args.extend(["--html", self.__html_path, "--self-contained-html"])

        self.__pytest_args.extend(pytest_report_args)
        self.__pytest_args.extend(self.__test_args)

        if only_enumerate:
            self.__pytest_args.append("--collect-only")

        print("Executing pytest with args: {}".format(self.__pytest_args))
        pytest.main(self.__pytest_args)



'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import requests
import json
import os

from arjuna.setu import Setu


class SetuAgentRequester:

    def __init__(self, base_url):
        self.base_url = base_url

    def __convert_response_text_to_json(self, response_text):
        try:
            return json.loads(response_text)
        except:
            raise Exception("Response text is not JSON. Raw response text: " + response_text)

    def __raise_exception_if_error(self, response_dict):
        if response_dict is None:
            raise Exception("Setu Actor's JSON response is null.")
        if (response_dict["result"].lower() == "error"):
            emessage = response_dict.get("emessage", "")
            raise Exception("Agent Error: " + emessage + os.linesep + "Trace: " + response_dict["etrace"])

    def get(self, url):
        req_url = self.base_url + url
        response = requests.get(req_url)
        response_json = json.loads(response.text)
        self.__raise_exception_if_error(response_json)
        return json.loads(response.text, encoding="utf-8")

    def post(self, url, json_dict):
        req_url = self.base_url + url
        response = requests.post(req_url, json=json_dict)
        response_json = self.__convert_response_text_to_json(response.text)
        self.__raise_exception_if_error(response_json)
        return json.loads(response.text)

    def get_base_url(self):
        return self.base_url


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import json
from enum import Enum

from arjuna.configure.impl.processor import ConfigCreator, CentralConfigLoader, ProjectConfigCreator
from arjuna.core.reader.hocon import HoconStringReader, HoconConfigDictReader
from arjuna.core.value import AnyRefValue

class TestConfigurator:

    def __init__(self, run_id):
        self.__run_id = run_id
        self.__default_ref_config = None
        self.__config_map = {}
        self.__cli_central_config = None
        self.__cli_test_config = None
        self.__project_config_loaded = False

    def __create_config_from_option_dicts(self, reference, arjuna_options, user_options):
        def format_value(val):
            if isinstance(val, AnyRefValue):
                return val.as_str()
            elif isinstance(val, Enum):
                return val.name
            else:
                return str(val)

        # ANCHOR
        crawdict = {
            "arjunaOptions": {k:format_value(v) for k,v in arjuna_options.items()},
            "userOptions": {k:format_value(v) for k,v in user_options.items()}
        }
        hreader = HoconStringReader(json.dumps(crawdict))
        hreader.process()

        config = ConfigCreator.create_new_conf(
            self.__default_ref_config.processor,
            reference,
            hreader.get_map()
        )
        return config

    def __init_cli_dicts(self, arjunaCentralOptions={}, arjunaTestOptions={}, userCentralOptions={}, userTestOptions={}):
        self.__cli_central_config = self.__create_config_from_option_dicts(None, arjunaCentralOptions, userCentralOptions)
        self.__cli_test_config = self.__create_config_from_option_dicts(None, arjunaTestOptions,
                                                                        userTestOptions)

    def init(self, root_dir, cli_config):
        self.__default_ref_config = CentralConfigLoader(root_dir, self.__run_id).config
        # self.__config_map[self.__default_ref_config.id] = self.__default_ref_config
        cli_config = cli_config and cli_config or {}
        self.__init_cli_dicts(**cli_config)
        return self.__default_ref_config

    def create_project_conf(self):
        project_conf_loader = ProjectConfigCreator(self.__default_ref_config)
        self.__default_ref_config = project_conf_loader.config
        self.__default_ref_config.process_arjuna_options()
        self.__default_ref_config.update(self.__cli_central_config)
        self.__default_ref_config.update(self.__cli_test_config)
        # self.__config_map[self.__default_ref_config.id] = self.__default_ref_config
        return self.__default_ref_config

    @property
    def ref_config(self):
        return self.__default_ref_config

    def get_config(self, id):
        return self.__config_map[id]

    # def get_arjuna_option_value(self, config_setu_id, option):
    #     sname = ArjunaOption[option.upper().strip().replace(".", "_")]
    #     rvalue = self.__config_map[config_setu_id].get_arjuna_option_value(sname)
    #     return rvalue

    # def get_user_option_value(self, config_setu_id, option):
    #     sname = option.upper().strip().replace(".", "_")
    #     rvalue = self.__config_map[config_setu_id].user_config.value(sname)
    #     return rvalue

    # def get_central_arjuna_option_value(self, option):
    #     sname = ArjunaOption[option.upper().strip().replace(".", "_")]
    #     rvalue = self.__default_ref_config.get_arjuna_option_value(sname)
    #     return rvalue

    # def get_central_user_option_value(self, option):
    #     sname = option.upper().strip().replace(".", "_")
    #     rvalue = self.__default_ref_config.user_config.value(sname)
    #     return rvalue

    def register_new_config(self, arjuna_options, user_options, parent_config=None):
        # Registering a config is post project conf registration. If no project conf, set it to true.
        self.__project_config_loaded = True
        reference = parent_config and parent_config._wrapped_config or self.__default_ref_config._wrapped_config
        config = self.__create_config_from_option_dicts(reference, arjuna_options, user_options)
        config.update(self.__cli_test_config)
        config.process_arjuna_options()
        return config


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

# Skeleton. needs to be ported.

class ArjunaSingleton:

    def __init__(self):
        self.__DEFAULT_CONTEXT_NAME = "default_context"
        self.__root_dir = None
        self.__test_session = test_session
        self.__ref_config = None
        self.__cli_config = None
        self.__logger = None
        self.__test_context_map = dict()

    def init(self, root_dir=None):
        self.__root_dir = root_dir
        self.__cli_config = CliArgsConfig()
        # Finalize logger
        createLogger("arjuna", rootDir + File.separator + "log" + File.separator + "arjuna-java.log")
        logger = Logger.getLogger("arjuna")
        Console.init()

        self.__test_session = DefaultTestSession()
        self.__ref_config = self.__test_session.init(root_dir)

        context = this.createTestContext(self.__DEFAULT_CONTEXT_NAME)
        self.__test_session[self.__DEFAULT_CONTEXT_NAME] = context
        return context

    @property
    def cli_config(self):
        return self.__cli_config

    @property
    def ref_config(self):
        return self.__ref_config

    @property
    def root_dir(self):
        return self.__root_dir

    def register_test_context(self, context):
        self.__test_context_map[context.name.lower()] = context

    def get_test_context(name):
        if name is None:
            raise Exception("Context name was passed as None.")
        try:
            return self.__test_context_map[name.lower()]
        except:
            raise Exception("No context found with name: " + name)

    def create_test_context(name):
        return DefaultTestContext(self.__test_session, name)

    def normalize_user_option(self, option):
        return option.strip().upper().replace(".", "_")

    def normalize_arjuna_option(self, option):
        return ArjunaOption[self.normalize_user_option(option)]

    def create_data_source_builder(self):
        return DataSourceBuilder(self.__test_session)

    def create_gui_automator(self, config=None, extended_config=None):
        return DefaultGuiAutomator(config and config or self.__ref_config, extended_config)

    def create_logger(self, logger_name, log_file_path):
        pass


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

'''
import logging
import sys
import inspect
import time

from arjuna.interact.gui.auto.automator.guiautomator import GuiAutomator
from arjuna.interact.gui.auto.finder.emd import GuiElementMetaData, SimpleGuiElementMetaData
from functools import partial

class _DummyLogger:

    def dummy(self, item, *vargs):
        print("Dummy call", item)
        frame = inspect.stack()[3]
        caller_str = "{}.py:{}:L{}".format(inspect.getmodule(frame[0]).__name__, frame[3], frame[2])
        print("Dummy logger called by {} with args: ".format(caller_str), str(vargs)[:300], "...")
        pass

    def __getattr__(self, item):
        return partial(self.dummy, item)


class Setu:
    LOGGER = None
    __DUMMY_LOGGER = _DummyLogger()

    @classmethod


    @classmethod
    def get_logger(cls):
        return cls.LOGGER and cls.LOGGER or cls.__DUMMY_LOGGER


__all__ = ["Setu", "GuiAutomator", "GuiElementMetaData", "SimpleGuiElementMetaData"]
'''


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''


from .driverelement import SeleniumDriverElementDispatcher
from arjuna.interact.gui.dispatcher.driver.driver_commands import DriverCommands
from arjuna.interact.gui.dispatcher.driver.element_finder import ElementFinder
from arjuna.interact.gui.dispatcher.driver.melement import MultiElement
from selenium.webdriver.remote.webelement import WebElement

class SeleniumDriverDispatcher:

    def __init__(self):
        self.__config = None
        self.__driver = None

    def __create_gui_element_dispatcher(self, element):
        return SeleniumDriverElementDispatcher.create_dispatcher(self, element)

    @property
    def driver(self):
        return self.__driver

    def launch(self, config):
        self.__config = config
        from .browser_launcher import BrowserLauncher
        self.__driver = BrowserLauncher.launch(config) 

    def quit(self):
        DriverCommands.quit(self.__driver)

    def go_to_url(self, url):
        DriverCommands.go_to_url(self.__driver, url)

    def go_back_in_browser(self):
        DriverCommands.go_back_in_browser(self.__driver)

    def go_forward_in_browser(self):
        DriverCommands.go_forward_in_browser(self.__driver)

    def refersh_browser(self):
        DriverCommands.refersh_browser(self.__driver)

    def get_source(self):
        return DriverCommands.get_source(self.__driver)

    def execute_javascript(self, script, *args):
        return DriverCommands.execute_javascript(self.__driver, script, 
                *[
                isinstance(arg, SeleniumDriverElementDispatcher) and arg.driver_element or arg for arg in args
            ]
        )

    def take_screenshot(self, file_path):
        DriverCommands.take_screenshot(self.__driver, file_path)

    def take_screenshot_as_base64(self):
        return DriverCommands.take_screenshot_as_base64(self.__driver)

    def find_element(self, with_type, with_value):
        element = ElementFinder.find_element(self.__driver, with_type, with_value)
        return 1, self.__create_gui_element_dispatcher(element)

    def __process_single_js_element(self, element):
        # JS returns null, undefined
        if element is None:
            raise Exception("JavaScript could not find element.")
        elif not isinstance(element, WebElement):
            raise Exception("JavaScript returned a non-element object.")
        else:
            return element

    def __process_js_element_list(self, elements):
        if not elements: raise Exception("JavaScript could not find element.")
        return [self.__process_single_js_element(e) for e in elements]

    def __process_js_element(self, element):
        if type(element) is list:
            element = self.__process_js_element_list(element)[0]
        else:
            element = self.__process_single_js_element(element)
        return element

    def __process_js_multielement(self, elements):
        if type(elements) is list:
            elements = self.__process_js_element_list(elements)
        else:
            elements = [self.__process_single_js_element(elements)]
        return elements

    def find_element_with_js(self, js):
        element = self.execute_javascript(js)
        element = self.__process_js_element(element)
        return 1, self.__create_gui_element_dispatcher(element)

    def find_multielement(self, with_type, with_value):
        web_elements = ElementFinder.find_elements(self.__driver, with_type, with_value)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_size(), melement

    def find_multielement_with_js(self, js):
        web_elements = self.execute_javascript(js)
        web_elements = self.__process_js_multielement(web_elements)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_size(), melement

    def get_current_window_handle(self):
        return DriverCommands.get_current_window_handle(self.__driver)

    def get_current_window_title(self):
        return DriverCommands.get_window_title(self.__driver)

    def maximize_current_window(self):
        DriverCommands.maximize_window(self.__driver)

    def get_current_window_size(self):
        res = DriverCommands.get_current_window_size(self.__driver)
        return {"width" : res[0], "height" : res[1]}

    def get_all_window_handles(self):
        return DriverCommands.get_all_winodw_handles(self.__driver)

    def focus_on_window(self, handle):
        DriverCommands.focus_on_window(self.__driver, handle)

    def close_current_window(self):
        DriverCommands.close_current_window(self.__driver)

    def is_web_alert_present(self):
        return DriverCommands.is_web_alert_present(self.__driver)

    def confirm_web_alert(self):
        DriverCommands.confirm_web_alert(self.__driver)

    def dismiss_web_alert(self):
        DriverCommands.dismiss_web_alert(self.__driver)

    def get_text_from_web_alert(self):
        return DriverCommands.get_text_from_web_alert(self.__driver)

    def send_text_to_web_alert(self, text):
        DriverCommands.send_text_to_web_alert(self.__driver, text)

    def focus_on_frame(self, element_dispatcher):
        DriverCommands.focus_on_frame(self.__driver, element_dispatcher.driver_element)

    def get_element_for_setu_id(self, id):
        return self.__driver_elements[id]

    def focus_on_parent_frame(self):
        DriverCommands.focus_on_parent_frame(self.__driver)

    def focus_on_dom_root(self):
        DriverCommands.focus_on_dom_root(self.__driver)

    def perform_action_chain(self, action_chain):
        DriverCommands.perform_action_chain(self, self.__driver, action_chain)

    def hover_on_element(self, element_dispatcher):
        DriverCommands.hover_on_element(self.__driver, element_dispatcher.driver_element)

    def mouse_click_on_element(self, element_dispatcher):
        DriverCommands.mouse_click_on_element(self.__driver, element_dispatcher.driver_element)

    def scroll_to_element(self, element_dispatcher):
        DriverCommands.scroll_to_element(self.__driver, element_dispatcher.driver_element)





'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''


import abc
import os
from enum import Enum
import unittest

from arjuna.interact.gui.auto.finder.emd import GuiElementMetaData

from .guidef import *

class AsserterMixIn:

    def __init__(self):
        # Trick to use assertions outside of a unittest test
        self._asserter = unittest.TestCase('__init__')

class Gui(AsserterMixIn):

    def __init__(self, *, config=None, ext_config=None, label=None):
        '''
            You can either provide automator.
        '''
        super().__init__()
        from arjuna import Arjuna
        self.__config = config is not None and config or Arjuna.get_ref_config()
        self.__econfig = ext_config
        if ext_config is None:
            self.__econfig = dict()
        else:
            if type(ext_config) is dict:
                self.__econfig = ext_config
            else:
                self.__econfig = ext_config.config
        self.__label = label is not None and label or self.__class__.__name__

    @property
    def config(self):
        return self.__config

    @property
    def ext_config(self):
        return self.__econfig

    @property
    def label(self):
        return self.__label

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def qual_name(self):
        return self.__class__.__qualname__


class AppContent(Gui):

    def __init__(self, *args, automator, label=None, **kwargs):
        super().__init__(config=automator.config, ext_config=automator.ext_config, label=label)
        self.__app = automator.app
        self.__automator = automator
        from arjuna import Arjuna
        self.__guimgr = Arjuna.get_gui_mgr()
        self.__guidef = None

        self.__gui_registered = False
        self.__def_file_name = None
        self.__def_file_path = None

    @property
    def app(self):
        return self.__app
    
    @property
    def automator(self):
        return self.__automator

    @property
    def gui_def(self):
        return self.__guidef

    def externalize_def(self, ns_dir=None, def_file_name=None):
        self.__def_file_name = def_file_name is not None and def_file_name or "{}.gns".format(self.label)        
        from arjuna.core.enums import ArjunaOption
        ns_dir = ns_dir and ns_dir or self.app.ns_dir
        ns_root_dir = self.config.get_arjuna_option_value(ArjunaOption.GUIAUTO_NAMESPACE_DIR).as_str()
        self.__def_file_path = os.path.join(ns_root_dir, ns_dir, self.def_file_name)
        self.__guidef = GuiDef(self.__guimgr.name_store, self.automator, self.label, self.__def_file_path) # self.__guimgr.namespace_dir, 
        # if register:
        #     self._register()

    @property
    def def_file_name(self):
        return self.__def_file_name

    @property
    def def_file_path(self):
        return self.__def_file_path

    def prepare(self):
        # Children can override and write any necessary preparation instructions e.g. externalizing
        pass

    def reach_until(self):
        # Children can override and write any necessary loading instructions
        pass

    def validate_readiness(self):
        pass

    def _load(self, *args, **kwargs):
        self.prepare(*args, **kwargs)
        try:
            self.validate_readiness()
        except:
            try:
                self.reach_until()
                self.validate_readiness()
            except Exception as e:
                raise Exception("GUI [{}] did not load as expected. Error: {}.".format(self.qual_name, str(e)))

    def transit(self, page):
        pass

    def convert_to_with_lmd(self, *raw_str_or_with_locators):
        from arjuna.interact.gui.helpers import With, WithType
        out = []
        for locator in raw_str_or_with_locators:
            w = None
            if isinstance(locator, With):
                w = locator
            elif type(locator) is str:
                w = With.gns_name(locator)
            elif isinstance(locator, Enum):
                w = With.gns_name(locator.name)
            else:
                raise Exception("A With object or name of element is expected as argument.")

            if w.wtype == WithType.GNS_NAME:
                if isinstance(w.wvalue, Enum):
                    w.wvalue = w.wvalue.name
                out.extend(self.gui_def.convert_to_with(w))
            else:
                out.append(w)
        lmd = GuiElementMetaData.create_lmd(*out)
        return lmd

    @property
    def browser(self):
        return self.impl_gui.browser

    def element(self, *str_or_with_locators, iconfig=None):
        return self.automator.element(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def multi_element(self, *str_or_with_locators, iconfig=None):
        return self.automator.multi_element(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def dropdown(self, *str_or_with_locators, option_container_locator=None, option_locator=None, iconfig=None):
        return self.automator.dropdown(
            self, 
            self.convert_to_with_lmd(*str_or_with_locators),
            option_container_lmd=option_container_locator and self.convert_to_with_lmd(option_container_locator) or None,
            option_lmd=option_locator and self.convert_to_with_lmd(option_locator) or None,
            iconfig=iconfig
        )

    def radio_group(self, *str_or_with_locators, iconfig=None):
        return self.automator.radio_group(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    def tab_group(self, *str_or_with_locators, tab_header_locator, content_relation_attr, content_relation_type, iconfig=None):
        return self.automator.tab_group(
            self,
            self.convert_to_with_lmd(*str_or_with_locators),
            tab_header_lmd=self.convert_to_with_lmd(tab_header_locator),
            content_relation_attr=content_relation_attr, 
            content_relation_type=content_relation_type, 
            iconfig=iconfig
        )

    def wait_until_element_absent(self, *str_or_with_locators):
        return self.automator.wait_until_element_absent(self.convert_to_with_lmd(*str_or_with_locators))

    @property
    def dom_root(self):
        return self.automator.dom_root(self)

    def frame(self, *str_or_with_locators, iconfig=None):
        return self.automator.frame(self, self.convert_to_with_lmd(*str_or_with_locators), iconfig=iconfig)

    @property
    def alert(self):
        return self.automator.alert

    @property
    def main_window(self):
        return self.automator.main_window

    def child_window(self, *str_or_with_locators):
        return self.automator.child_window(self.convert_to_with_lmd(*str_or_with_locators))

    @property
    def latest_child_window(self):
        return self.automator.latest_child_window

    def close_all_child_windows(self):
        self.automator.close_all_child_windows()

    @property
    def browser(self):
        return self.automator.browser

    def set_slomo(self, on, interval=None):
        self.automator.set_slomo(on, interval)

    def execute_javascript(self, js, *args):
        return self.automator.execute_javascript(js, *args)

    def take_screenshot(self, name=None):
        return self.automator.take_screenshot(name)





'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

class PytestHooks:


    @classmethod
    def get_request_attr(cls, item, obj_name):
        request =  cls.get_request_obj(item)
        return getattr(cls.get_container_based_on_scope(request), obj_name)

    @classmethod
    def get_request_obj(cls, item):
        return item.funcargs['request']

    @classmethod
    def get_plugin(cls, item, name):
        return item.config.pluginmanager.getplugin(name)

    @classmethod
    def get_html_report_plugin(cls, item):
        return cls.get_plugin(item, 'html')

    @classmethod
    def add_screenshot_for_failed_result(cls, html_plugin, outcome, *, screenshoter=None):
        pytest_html = html_plugin
        report = outcome.get_result()
        extra = getattr(report, 'extra', [])
        if report.when == 'call':

            xfail = hasattr(report, 'wasxfail')
            if (report.skipped and xfail) or (report.failed and not xfail):
                # extra.append(pytest_html.extras.url(app.base_url))
                fpath, fb64 = screenshoter.automator.take_screenshot()
                img_elem = '''<img src="data:image/png;base64,{}"/>'''.format(fb64)
                extra.append(
                    pytest_html.extras.html(
                        '''<div class="image"><a href="{}" target="_blank">{}</a>'''.format(fpath, img_elem)
                    )
                )
                report.extra = extra


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import re
from enum import Enum, auto

class ConfigPropertyType(Enum):
	ARO = auto()
	ACO = auto()
	URO = auto()
	UCO = auto()

class CliArgsConfig:
    ARJ_PATTERN = re.compile("(?i)(aro|aco)\\s*:\\s*(\\S+)\\s*")
    USER_PATTERN = re.compile("(?i)(uro|uco)\\s*:\\s*(\\S+)\\s*")

    def __init__(self, options):
        self.__arj_options_map = {
            ConfigPropertyType.ARO : dict(),
            ConfigPropertyType.ACO : dict()
        }
        self.__user_options_map = {
            ConfigPropertyType.URO : dict(),
            ConfigPropertyType.UCO : dict()
        }

        for option in options:
            m = CliArgsConfig.ARJ_PATTERN.match(option.name)
            if m is not None:
                c_type = ConfigPropertyType[m.group(1).upper()]
                try:
                    arj_prop = m.group(2)
                except:
                    raise Exception("Empty Arjuna Option Key provided in CLI: {}".format(option.name))
                else:
                    try:
                        arj_option = Arjuna.normalise_arjuna_option(arj_prop)
                    except:
                        raise Exception("You have provided an invalid Arjuna Option in CLI: {}".format(option.name))
                    else:
                        self.__arj_options_map[c_type][arj_option] = option.value
                        continue

            m = CliArgsConfig.USER_PATTERN.match(option.name)
            if m is not None:
                c_type = ConfigPropertyType[m.group(1).upper()]
                try:
                    user_option = Arjuna.normalise_user_option(m.group(2))
                except:
                    raise Exception("Empty User Option key provided in CLI: {}".format(option.name))
                else:
                    self.__arj_options_map[c_type][user_option] = option.value
                    continue

    def as_map(self):
        map = dict()
        map["arjuna_reference_options"] = self.__get_arjuna_reference_options()
        map["arjuna_custom_options"] = self.__get_arjuna_custom_options()
        map["user_reference_options"] = self.__get_user_reference_options()
        map["user_reference_options"] = self.__get_user_custom_options()

    def __get_arjuna_reference_options(self):
        return self.__arj_options_map[ConfigPropertyType.ARO]

    def __get_arjuna_custom_options(self):
        return self.__arj_options_map[ConfigPropertyType.ACO]

    def __get_user_reference_options(self):
        return self.__arj_options_map[ConfigPropertyType.URO]

    def __get_user_custom_options(self):
        return self.__arj_options_map[ConfigPropertyType.UCO]


#### From old Setu code
from arjuna.core.enums import *
from arjuna.core.value import AnyRefValue


class DefaultTestConfig:

    def __init__(self, test_session, name, config):
        super().__init__()
        self.__session = test_session
        self.__name = name
        self.__wrapped_config = config

    @property
    def _wrapped_config(self):
        return self.__wrapped_config

    @property
    def test_session(self):
        return self.__session

    def __fetch_config_option_value(self, setu_action_type, option_str):
        response = self._send_request(setu_action_type, SetuArg.arg("option", option_str))
        return response.get_value()

    @staticmethod
    def normalize_option_str(option_str):
        option_str = isinstance(option_str, Enum) and option_str.name or option_str
        return option_str.upper().strip().replace(".", "_")

    @staticmethod
    def normalize_arjuna_option_str(option_str):
        return ArjunaOption[DefaultTestConfig.normalize_option_str(option_str)]

    def get_arjuna_option_value(self, option):
        arjuna_option = option
        if type(option) is str:
            arjuna_option = DefaultTestConfig.normalize_arjuna_option_str(option)
        return AnyRefValue(self.__wrapped_config.arjuna_config.value(arjuna_option))

    def get_user_option_value(self, option):
        user_option = DefaultTestConfig.normalize_option_str(option)
        return AnyRefValue(self.__wrapped_config.user_config.value(user_option))

    def get_arjuna_options_as_map(self):
        return self.__wrapped_config.arjuna_config.as_json_dict()

    def is_arjuna_option_not_set(self, option):
        return self.__wrapped_config.arjuna_config.is_not_set(option)

    @property
    def name(self):
        return self.__name

    @property
    def guiauto_context(self):
        return self.get_arjuna_option_value(ArjunaOption.GUIAUTO_CONTEXT).as_enum(GuiAutomationContext)

    @property
    def browser_type(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_NAME).as_enum(BrowserName)

    @property
    def browser_version(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_VERSION).as_str()

    @property
    def browser_binary_path(self):
        return self.get_arjuna_option_value(ArjunaOption.BROWSER_BIN_PATH).as_str()

    @property
    def test_run_env_name(self):
        return self.get_arjuna_option_value(ArjunaOption.TESTRUN_ENVIRONMENT).as_str()

    @property
    def screenshots_dir(self):
        return self.get_arjuna_option_value(ArjunaOption.SCREENSHOTS_DIR).as_str()

    @property
    def log_dir(self):
        return self.get_arjuna_option_value(ArjunaOption.LOG_DIR).as_str()

    @property
    def guiauto_max_wait_time(self):
        return self.get_arjuna_option_value(ArjunaOption.GUIAUTO_MAX_WAIT).as_int()

    def as_map(self):
        return self.__wrapped_config.as_json_dict()

class CliArgsConfig:

    def __init__(self, arg_dict):
        self.__aco = {}
        self.__ato = {}
        self.__uco = {}
        self.__uto = {}

        kinds = {
            "aco": self.__aco,
            "ato": self.__ato,
            "uco": self.__uco,
            "uto": self.__uto
        }

        lower_actual_key_map = {i.lower():i for i in arg_dict}
        for kind in kinds:
            if kind in lower_actual_key_map:
                actual_key = lower_actual_key_map[kind]
                d_item = arg_dict[actual_key]
                if d_item:
                    for entry in d_item:
                        k,v = entry
                        kinds[kind][k.lower()] = v
                del arg_dict[actual_key]

        for akey, avalue in arg_dict.items():
            self.__aco[akey.lower()] = avalue

    def as_map(self):
        return {
            "arjunaCentralOptions": self.__aco,
            "arjunaTestOptions": self.__ato,
            "userCentralOptions": self.__uco,
            "userTestOptions": self.__uto
        }



'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import uuid
from arjuna.interact.gui.gom.namestore import GuiNameStore
from arjuna.core.enums import ArjunaOption

class GuiManager:

    def __init__(self, config):
        self.__name_store = GuiNameStore()
        self.__namespace_dir = config.arjuna_config.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)

    @property
    def name_store(self):
        return self.__name_store

    @property
    def namespace_dir(self):
        return self.__namespace_dir

    def define_gui(self, automator, label=None, name=None, qual_name=None, def_file_path=None):
        from arjuna.interact.gui.gom.guidef import GuiDef
        gui_def = GuiDef(self.__name_store, self.__namespace_dir, automator, label, def_file_path)
        gui = Gui(self, automator, gui_def)
        return gui





'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os

from arjuna import Arjuna
from arjuna.interact.gui.auto.element.guielement import GuiElement
from arjuna.interact.gui.auto.finder.emd import SimpleGuiElementMetaData
from arjuna.interact.gui.auto.source.parser import FrameSource
from arjuna.interact.gui.auto.base.configurable import Configurable

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.core.exceptions import ChildWindowNotFoundError

class FrameConditions:

    def __init__(self, frame):
        self.__frame = frame

    @property
    def frame(self):
        return self.__frame

    def FrameIsPresent(self, lmd, *args, **kwargs):
        caller = DynamicCaller(self.frame._find_frame, lmd, *args, **kwargs)
        return CommandCondition(caller)

class FrameContainer(Configurable):

    def __init__(self, gui, iconfig=None):
        super().__init__(gui, iconfig)
        self.__gui = gui
        self.__automator = gui.automator
        self.__conditions = FrameConditions(self)

    @property
    def conditions(self):
        return self.__conditions

    @property
    def gui(self):
        return self.__gui

    @property
    def automator(self):
        return self.__automator

    @property
    def max_wait_time(self):
        return self.automator.config.guiauto_max_wait_time

    def __check_tag(self, wrapped_element):
        tag = wrapped_element.source.tag
        if tag.lower() != "iframe":
            raise Exception("The element should have a 'iframe' tag for IFrame element. Found: " + tag)

    def frame(self, locator_meta_data, iconfig=None):
        return self.conditions.FrameIsPresent(locator_meta_data, iconfig=iconfig).wait(max_wait_time=self.max_wait_time)

    def _find_frame(self, locator_meta_data, iconfig=None):
        iconfig = iconfig and iconfig or self.settings
        found = False
        frame = None
        for locator in locator_meta_data.locators: 
            try:
                if locator.ltype.name == "INDEX":
                    index = locator.lvalue
                    emd = SimpleGuiElementMetaData("xpath", "//iframe")
                    multi_element = self.automator.multi_element(self.gui, emd)
                    # multi_element.find()
                    try:
                        wrapped_element = multi_element[index]
                    except:
                        # In case another identifier is present it should be tried.
                        continue
                    self.__check_tag(wrapped_element)
                    frame = IPartialFrame(self.gui, self, multi_element, wrapped_element, iconfig=iconfig)
                else:
                    emd = SimpleGuiElementMetaData(locator.ltype.name, locator.lvalue)
                    wrapped_element = self.automator.element(self.gui, emd)
                    # wrapped_element.find()
                    self.__check_tag(wrapped_element)
                    frame = IFrame(self.gui, self, wrapped_element, iconfig=iconfig)

                return frame
            except WaitableError as f:
                continue  
            except Exception as e:
                raise Exception(str(e) + traceback.format_exc()) 

        raise ChildFrameNotFoundError(*locator_meta_data.locators)

    def enumerate_frames(self):
        self.focus()
        emd = SimpleGuiElementMetaData("xpath", "//iframe")
        multi_element = self.automator.create_multielement(emd)
        ret_str = os.linesep.join(["--> " + s for s in multi_element.get_source()._get_root_content_as_list()])
        return self._source_parser.get_root_content() + os.linesep + ret_str

class DomRoot(FrameContainer):

    def __init__(self, gui):
        super().__init__(gui)
        self.__frame_context = "root"
        self._source_parser = self.automator.source

    @property
    def frame_context(self):
        return self.__frame_context

    def __set_frame_context_str(self, name):
        self.__frame_context = name
        Arjuna.get_logger().debug("Automator is in {} frame".format(self.__frame_context))
        print("Automator is in {} frame".format(self.__frame_context))

    def set_frame_context(self, frame):
        self.__set_frame_context_str(frame)

    def set_frame_context_as_root(self):
        self.__set_frame_context_str("root")  

    def is_in_root_context(self):
        return self.__frame_context == "root"

    def focus(self):
        self.automator.dispatcher.focus_on_dom_root()
        self.set_frame_context_as_root() 

    @property
    def get_source(self):
        self.focus()
        return self.automator.source

class IFrame(FrameContainer):

    def __init__(self, gui, dom_root, wrapped_element, iconfig=None):
        super().__init__(gui, iconfig=iconfig)
        self.__dom_root = dom_root
        self.__parent_frames = []
        self.__wrapped_element = wrapped_element
        self._source_parser = FrameSource(self)

    @property
    def dom_root(self):
        return self.__dom_root

    @property
    def wrapped_element(self):
        return self.__wrapped_element

    def set_parents(self, parents):
        self.__parent_frames = parents

    def _act(self, json_dict):
        return self.dom_root._act(json_dict)

    def _focus_on_parents(self):
        if self.__parent_frames:
            for parent in self.__parent_frames:
                parent.focus()

    def __reload_wrapped_element(self):
        self.wrapped_element.find()

    def focus(self):
        if not self.dom_root.is_in_root_context():
            self.dom_root.focus()
            self._focus_on_parents()
        # self.wrapped_element.find()
        self._source_parser.set_root_source(self.wrapped_element.source.content.root)
        self.automator.dispatcher.focus_on_frame(self.wrapped_element.dispatcher)
        self.dom_root.set_frame_context(self)

    def _get_html_content_from_remote(self):
        return self.automator.get_source()

    def get_wrapped_element(self):
        return self.wrapped_element

    @property
    def source(self):
        self._source_parser.load()
        return self._source_parser

    # def focus_on_parent(self):
    #     self._act(TestAutomatorActionBodyCreator.jump_to_parent_frame())
    #     if self.__parent_frames:
    #         self.dom_root.set_frame_context(self.__parent_frames[-1])
    #     else:
    #         self.dom_root.set_frame_context_as_root()

    @property
    def parent(self):
        if self.__parent_frames:
            return self.__parent_frames[-1]
        else:
            return self.dom_root


class IPartialFrame(IFrame):

    def __init__(self, gui, dom_root, melement, wrapped_element, iconfig=None):
        super().__init__(gui, dom_root, wrapped_element, iconfig=iconfig)
        self.__melement = melement

    def focus(self):
        # self.__melement.find()
        self.automator.dispatcher.focus_on_frame(self.wrapped_element.dispatcher)
        self.dom_root.set_frame_context(self)


'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.core.enums import ArjunaOption
from arjuna.client.core.action import *
from arjuna.interact.gui.auto.finder.emd import SimpleGuiElementMetaData

from .guiautomator import GuiAutomator
from .handler import Handler

class SingleActionChain(Handler):

    def __init__(self, automator, element=None):
        super().__init__(automator)
        from arjuna.interact.gui.auto.element.frame import DomRoot
        self.__automator = automator
        self.__element = element
        self.__attached_to_element = element is not None

    def is_click_action(self, action):
        return PartialActionType[action["actionType"]] == PartialActionType.CLICK

    def are_args_defined(self, action):
        return "args" in action

    def get_args(self, action):
        return action["args"]

    def contains_target_element(self, action):
        return "targetElement" in self.get_args(action)

    def get_target_element(self, action):
        return action["args"]["targetElement"]

    def contains_target_point(self, action):
        return "targetPoint" in self.get_args(action)

    def get_target_point(self, action):
        return action["args"]["targetPoint"]

    def perform(self, single_action_chain):
        processed_list = []
        current_action_list = []
        for action in single_action_chain:
            if self.is_click_action(action):
                if not self.are_args_defined():
                    current_action_list.append(("click", dict()))
                    continue

                target_element = None
                if self.contains_target_element(action):
                    target_element_id = self.get_target_element(action)
                    target_element = self.__automator.get_element_for_setu_id(target_element_id)
                elif self.contains_target_point(action):
                    emd = SimpleGuiElementMetaData("point", self.get_target_point(action))
                    element = self.__automator.create_element(emd)
                
                if target_element:
                    target_element.find_if_not_found()
                    current_action_list.append(("click", {"on_element" : (target_element.setu_id, True)}))
                    
        processed_list.append(current_action_list)
        for single_chain in processed_list:
            self.__automator.dispatcher.perform_action_chain(single_chain)


        






'''
This file is a part of Arjuna
Copyright 2015-2020 Rahul Verma

Website: www.RahulVerma.net

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
from .source import *

def create_file_data_source(data_dir, file_name, record_format="MAP", delimiter="\t"):
    ds = None
    ext = file_name.lower()
    path = os.path.join(data_dir, file_name)
    rformat = record_format.upper()
    if ext.endswith(".csv") or ext.endswith(".txt"):
        if rformat == "LIST":
            ds = DsvFileListDataSource(path, delimiter)
        else:
            ds = DsvFileMapDataSource(path, delimiter)
    elif ext.endswith(".xls"):
        if rformat == "LIST":
            ds = ExcelFileListDataSource(path)
        else:
            ds = ExcelFileMapDataSource(path)
    elif ext.endswith(".ini"):
        ds = IniFileDataSource(path)
    else:
        raise Exception("This is not a default file format supported as a data source: " + path)
    return ds
